#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <PubSubClientTools.h>
#include "MQ135.h"              // https://github.com/GeorgK/MQ135

#include <Thread.h>             // https://github.com/ivanseidel/ArduinoThread
#include <ThreadController.h>

#define WIFI_SSID "xxxxxxxxxx"        //add your SSID
#define WIFI_PASS "xxxxxxxxxx"        //add your Password

#define MQTT_SERVER "192.168.1.xxx" //add the IP of your MQTT brocker
#define MQTT_PORT 1883              //add port of your MQTT brocker. Usualy 1833. If not put yours
#define mqtt_user "xyz"             //add username of your MQTT Brocker
#define mqtt_password "xwz"         //add password of your MQTT Brocker

// topic definitions
#define RZero_topic "sensor/RZero"
#define RZero_Corrected_topic "sensor/RZero_Corrected"
#define Resistance_topic "sensor/Resistance"
#define PPM_topic "sensor/PPM"
#define PPM_Corrected_topic "sensor/PPM_Corrected"
#define CO2_topic "sensor/CO2"
#define Voltage_topic "sensor/Voltage"
#define Temp_Ext_topic "Room1/ext/Temp_Ext"
#define Humi_Ext_topic "Room1/ext/Humi_Ext"

// The load resistance on the board
#define RLOAD 10.0

// MQ135 sensor
#define PIN_MQ135 A0
MQ135 mq135_sensor = MQ135(PIN_MQ135);

WiFiClient espClient;
PubSubClient client(MQTT_SERVER, MQTT_PORT, espClient);
PubSubClientTools mqtt(client);
ThreadController threadControl = ThreadController();
Thread thread = Thread();

int value = 0;
const String s = "";
const int gasSensor =0;

// Temperature/Humidity receivet via MQTT from external sensor
String t = Temp_Ext_topic; 
String h = Humi_Ext_topic;

void setup() {
  Serial.begin(115200);
  Serial.println();
  // Connect to WiFi
  Serial.print(s+"Connecting to WiFi: "+WIFI_SSID+" ");
  WiFi.begin(WIFI_SSID, WIFI_PASS);
  while (WiFi.status() != WL_CONNECTED) {
      reconnect();
    delay(500);
    Serial.print(".");
  }
  Serial.println("connected");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  // Connect to MQTT
  Serial.print(s+"Connecting to MQTT: "+MQTT_SERVER+" ... ");
  if (client.connect("GasSensor")) {
    Serial.println("connected");

    mqtt.subscribe("Room1/ext/Temp_Ext", topic1_subscriber);
    mqtt.subscribe("Room1/ext/Humi_Ext", topic2_subscriber);
    /*      //Another available topic for future use
    //mqtt.subscribe("test_in/#",        topic3_subscriber);
    */
  } else {
    Serial.println(s+"failed, rc="+client.state());
  }
}

void reconnect() {
  // Loop until we're reconnected
  while (!client.connected()) {
    Serial.print("Attempting MQTT connection...");
    // Attempt to connect
    if (client.connect("GasSensor", mqtt_user, mqtt_password)) {

      Serial.println("connected");
    } else  {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
}
bool checkBound(float newValue, float prevValue, float maxDiff) {
  return !isnan(newValue) &&
         (newValue < prevValue - maxDiff || newValue > prevValue + maxDiff);
         client.subscribe("Temp_Ext_topic");
         client.subscribe("Humi_Ext_topic");
}

long lastMsg = 0;
float RZero = 0.0;
float RZero_corrected = 0.0;
float Resistance = 0.0;
float PPM = 0.0;
float PPM_corrected = 0.0;
float Voltage = 0.0;
float diff = 1.0;
float temperature = t.toFloat();
float humidity = h.toFloat();

void loop() {
  if (!client.connected() && WiFi.status() == 3) {reconnect();}
  client.loop();
  threadControl.run();

    long now = millis();
  if (now - lastMsg > 5000) {
    lastMsg = now;
// let's take the maximum info provided by MQ135 sensor
    float newVoltage = getVoltage(gasSensor);    //will be later defined
    float newRZero = mq135_sensor.getRZero();
    float newRZero_corrected = mq135_sensor.getCorrectedRZero(temperature, humidity); //corrects RZero acording to temperature & humidity
    float newResistance = mq135_sensor.getResistance();
    float newPPM = mq135_sensor.getPPM()*1000;
    float newPPM_corrected = mq135_sensor.getCorrectedPPM(temperature, humidity)*1000;  //corrects PPM acording to temperature & humidity

    if (checkBound(newVoltage, Voltage, diff)) {
      Voltage = newVoltage;
        //Print the values to serial
      Serial.print("New Voltage:");
      Serial.println(String(Voltage).c_str());
        //send the values to MQTT server
      client.publish(Voltage_topic, String(Voltage).c_str(), true);
    }
    if (checkBound(newRZero, RZero, diff)) {
      RZero = newRZero;
      Serial.print("New RZero:");
      Serial.println(String(RZero).c_str());
      client.publish(RZero_topic, String(RZero).c_str(), true);
    }
    if (checkBound(newRZero_corrected , RZero_corrected , diff)) {
      RZero_corrected  = newRZero_corrected ;
      Serial.print("New RZero_Corrected :");
      Serial.println(String(RZero_corrected).c_str());
      client.publish(RZero_Corrected_topic, String(RZero_corrected).c_str(), true);
    }
    if (checkBound(newResistance , Resistance , diff)) {
      Resistance  = newResistance ;
      Serial.print("New Resistance :");
      Serial.println(String(Resistance).c_str());
      client.publish(Resistance_topic, String(Resistance).c_str(), true);
    }
    if (checkBound(newPPM , PPM , diff)) {
      PPM  = newPPM ;
      Serial.print("New PPM :");
      Serial.println(String(PPM).c_str());
      client.publish(PPM_topic, String(PPM).c_str(), true);
    }
    if (checkBound(newPPM_corrected , PPM_corrected , diff)) {
      PPM_corrected  = newPPM_corrected ;
      Serial.print("New PPM_Corrected :");
      Serial.println(String(PPM_corrected).c_str());
      client.publish(PPM_Corrected_topic, String(PPM_corrected).c_str(), true);
    }
  }
}

float getVoltage(int pin){
  return (analogRead(pin) * 0.004882814 * 1000);
  // This equation converts the 0 to 1023 value that analogRead()
  // returns, into a 0.0 to 5.0 value that is the true voltage
  // being read at that pin.
}

void topic1_subscriber(String topic, String message_t) {
  Serial.println(s+"Message arrived in function 1 ["+topic+"] "+message_t);
  client.publish("ext_sensor/Temperature_Ext", String(message_t).c_str(),true);
  Serial.println(s+"Message sended in topic ext_sensor/Temperature_Ext: "+message_t);
}

void topic2_subscriber(String topic, String message_h) {
  Serial.println(s+"Message arrived in function 2 ["+topic+"] "+message_h);
  client.publish("ext_sensor/Humidity_Ext", String(message_h).c_str(),true);            
  Serial.println(s+"Message sended in topic ext_sensor/Humidity_Ext: "+message_h);
}

/*    //Another available topic for future use
void topic3_subscriber(String topic, String message) {
  Serial.println(s+"Message arrived in function 3 ["+topic+"] "+message);
}*/
